!
! first.inc  -  constants for assembler module for DOS boot loader
!
! Copyright (C) 1996-2001 Gero Kuhlmann   <gero@gkminix.han.de>
!
!  This program is free software; you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation; either version 2 of the License, or
!  any later version.
!
!  This program is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with this program; if not, write to the Free Software
!  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
! $Id: first.inc,v 1.2 2001/08/27 08:40:16 gkminix Exp $
!
!====================================================================
!
! Definitions for accessing the BIOS data area
!
BIOS_SEG	equ	$0040		! Segment of BIOS data area
BIOS_FDSTAT	equ	$0041		! Floppy disk status byte
BIOS_HDSTAT	equ	$0074		! Hard disk status byte

I13_INT		equ	$13 * 4		! interrupt vector 13h
I2F_INT		equ	$2F * 4		! interrupt vector 2Fh
IF1_INT		equ	$F1 * 4		! interrupt vector F1h
IF8_INT		equ	$F8 * 4		! interrupt vector F8h



!
!====================================================================
!
! Layout of disk boot sector
!
SECT_SIZE	equ	512		! sector size must be 512 bytes

DISK_BPS	equ	11		! offset of bytes per sectors
DISK_SECTS	equ	19		! offset of total number of sects
DISK_SPT	equ	24		! offset of sectors per track
DISK_HEADS	equ	26		! offset of number of heads
DISK_BOOT	equ	36		! offset of disk ID to boot from

PART_STATUS	equ	$01BE		! offset of partition status
PART_FRST_HEAD	equ	$01BF		! offset of first head number
PART_FRST_SECT	equ	$01C0		! offset of first sector number
PART_FRST_CYL	equ	$01C1		! offset of first cylinder number
PART_TYPE	equ	$01C2		! offset of partition id
PART_LAST_HEAD	equ	$01C3		! offset of last head number
PART_LAST_SECT	equ	$01C4		! offset of last sector number
PART_LAST_CYL	equ	$01C5		! offset of last cylinder number
PART_ABS_SECT	equ	$01C6		! offset of first sector number
PART_SEC_NUM	equ	$01CA		! offset of number of sectors

PART_ACTIVE	equ	$80		! indicates active partition
PART_FAT12	equ	$01		! indicates partition type
PART_FAT16	equ	$04		! indicates partition type

BOOT_ID_FD	equ	$00		! BIOS id of floppy boot disk
BOOT_ID_HD	equ	$80		! BIOS id of hard boot disk
BOOT_OFFSET	equ	$7C00		! offset for boot block in segment 0
TEMP_SEGMENT	equ	$3000		! segment for temporary storage



!
!====================================================================
!
! Layout of BIOS parameter block as used by DOS
!
BPB_BPS		equ	0		! offset to bytes per sector
BPB_SPC		equ	2		! offset to sectors per cluster
BPB_RES		equ	3		! offset to # of reserved sectors
BPB_FATS	equ	5		! offset to # of FATs
BPB_DIR		equ	6		! offset to # of root dir entries
BPB_TOT_SECTS	equ	8		! offset to total # of sectors
BPB_MEDIA_ID	equ	10		! offset to media ID
BPB_SPF		equ	11		! offset to # of sectors per FAT
BPB_SPT		equ	13		! offset to # of sectors per track
BPB_HEADS	equ	15		! offset to # of heads
BPB_HIDDEN	equ	17		! offset to # of hidden sectors



!
!====================================================================
!
! Flags for DOS drive parameter block
!
DPB_F_FIXED	equ	$0001		! fixed media
DPB_F_DOOR	equ	$0002		! drive supports door lock status
DPB_F_TSIZE	equ	$0004		! all sectors in track are same size
DPB_F_MULTI	equ	$0008		! multiple logical units in drive
DPB_F_SHARED	equ	$0010		! logical for shared physical drive
DPB_F_CHANGE	equ	$0020		! disk change detected
DPB_F_NEWPARA	equ	$0040		! device parameters were changed
DPB_F_FORMAT	equ	$0080		! disk reformatted
DPB_F_ACCESS	equ	$0100		! access flag - fixed media only

DPB_F_DEFAULT	equ	DPB_F_CHANGE + DPB_F_NEWPARA

DPB_T_360	equ	0		! type ID for 360kB disk
DPB_T_1200	equ	1		! type ID for 1.2MB disk
DPB_T_720	equ	2		! type ID for 720kB disk
DPB_T_SD8	equ	3		! type ID for SD 8 inch disk
DPB_T_HD8	equ	4		! type ID for HD 8 inch disk
DPB_T_FIXED	equ	5		! type ID for fixed disk
DPB_T_TAPE	equ	6		! type ID for tape drive
DPB_T_1440	equ	7		! type ID for 1.44MB disk
DPB_T_OPTICAL	equ	8		! type ID for optical disk
DPB_T_2880	equ	9		! type ID for 2.88MB disk



!
!====================================================================
!
! Vendor information for the boot rom image. These values have to be
! identical to those in mknbi.h.
!
#define VENDOR_MAGIC	"GK-mknbi-DOS"	! vendor ID
#define VENDOR_SIZE	3		! size of vendor ID in dwords

VENDOR_BOOTL	equ	16		! tag for boot loader segment
VENDOR_RAMDISK	equ	17		! tag for ramdisk image



!
!====================================================================
!
! Layout and magic ID of boot rom image header. These values have to
! be identical to those used by the boot rom. See SPEC.DOC of the
! boot rom source for further information.
!
BOOT_MAGIC	equ	$1B031336	! boot image header magic cookie
BOOT_SIZE	equ	512		! total size of boot image header

BOOT_HD_SIZE	equ	4		! size of header in dwords
BOOT_HD_MAGIC	equ	0		! offset for header magic number
BOOT_HD_LENGTH	equ	4		! offset for header length
BOOT_HD_FLAG1	equ	5		! offset for header flag 1
BOOT_HD_FLAG2	equ	6		! offset for header flag 2
BOOT_HD_FLAG3	equ	7		! offset for header flag 3
BOOT_HD_LOCN	equ	8		! offset for header location
BOOT_HD_EXEC	equ	12		! offset for execute address
BOOT_HD_VENDOR	equ	16		! offset for header vendor information

BOOT_LD_SIZE	equ	4		! size of load record in dwords
BOOT_LD_LENGTH	equ	0		! offset for load record length
BOOT_LD_TAG1	equ	1		! offset for load record tag 1
BOOT_LD_TAG2	equ	2		! offset for load record tag 2
BOOT_LD_FLAGS	equ	3		! offset for load record flags
BOOT_LD_ADDR	equ	4		! offset for absolute address
BOOT_LD_ILENGTH	equ	8		! offset for image length
BOOT_LD_MLENGTH	equ	12		! offset for memory length
BOOT_LD_VENDOR	equ	16		! offset for vendor information

BOOT_FLAG_B0	equ	$01		! mask for load record flag B0
BOOT_FLAG_B1	equ	$02		! mask for load record flag B1
BOOT_FLAG_EOF	equ	$04		! mask for load record flag EOF

BOOT_LD_SECNUM	equ	0		! offset of total number of sectors
BOOT_LD_SPT	equ	2		! offset of sectors per track
BOOT_LD_CYL	equ	4		! offset of number of cylinders
BOOT_LD_DRIVE	equ	6		! offset of boot drive ID
BOOT_LD_NOHD	equ	7		! offset of no-hard-disk flag



!
!====================================================================
!
! Layout and magic ID of bootp record. Refer to RFC 951, RFC 1048
! and RFC 1533 for further information. The BOOTP record can be
! longer than the standard length as the bootrom might be able to
! load a vendor extension file via tftp.
!
#define BOOTP_MAGIC_RFC	$63, $82, $53, $63	! RFC 1048 vendor ID
#define BOOTP_MAGIC_CMU	$43, $4D, $55, $00	! CMU vendor ID
#define BOOTP_MAGIC_STA	$53, $54, $41, $4E	! Stanford vendor ID
#define BOOTP_MAGIC_LEN	4			! length of vendor ID

BOOTP_REQUEST	equ	1		! bootp OP-code
BOOTP_REPLY	equ	2

BOOTP_OP	equ	0		! offset to bootp OP-code
BOOTP_VEND	equ	236		! offset to vendor information
BOOTP_SIZE	equ	300		! size of complete bootp record

BOOTP_RFC_NOP	equ	0		! RFC vendor tag for NO-OP
BOOTP_RFC_END	equ	255		! RFC vendor tag for end of record

