! first.S  -  primary boot loader
!
! Copyright (C) 1995-2001 Gero Kuhlmann   <gero@gkminix.han.de>
! Copyright (C) 1996-2001 Gero Kuhlmann   <gero@gkminix.han.de>
!                and Markus Gutschke <gutschk@math.uni-muenster.de>
!
!  This program is free software; you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation; either version 2 of the License, or
!  any later version.
!
!  This program is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with this program; if not, write to the Free Software
!  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
! $Id: first.S,v 1.2 2001/08/27 08:40:16 gkminix Exp $
!
!====================================================================
!
! We use 16-bit 386 code here, since Linux can only run on 386+
! processors.
!
use16	386


#ifndef ASM_DEBUG
#undef ASM_DEBUG
#endif
#include "first.inc"
#include <version.h>


	.text

	.org	0

! First we save the registers used by the bootrom to allow a return to
! it lateron if the boot fails for some reason. Since we havent checked
! for a 386+ yet we can only use plain 16-bit opcodes here.

	mov	dx,ds
	mov	ax,cs			! set DS and ES
	mov	ds,ax
	mov	[oldES],es
	mov	[oldDS],dx		! save old register values in case
	mov	[oldBP],bp		! we have to return to the boot rom
	mov	[oldSI],si
	mov	[oldDI],di
	mov	bp,sp
	mov	ax,[bp + 4]
	mov	[header + 0],ax		! load the address of the boot header
	mov	ax,[bp + 6]
	mov	[header + 2],ax
	mov	ax,[bp + 8]
	mov	[bootp + 0],ax		! load the address of the bootp block
	mov	ax,[bp + 10]
	mov	[bootp + 2],ax

! Tell the user who we are and that we started running

	mov	si,#sigmsg
	call	prnstr

! Show program parameter addresses for debugging

#if defined(ASM_DEBUG) || defined(ASM_DEBUG_VERBOSE)
	mov	si,#ad1msg
	call	prnstr
	mov	ax,[header + 2]
	call	prnwrd
	mov	al,#$3A
	call	prnchr
	mov	ax,[header + 0]		! print header address
	call	prnwrd
	mov	si,#crlf
	call	prnstr

	mov	si,#ad2msg
	call	prnstr
	mov	ax,[bootp + 2]
	call	prnwrd
	mov	al,#$3A
	call	prnchr
	mov	ax,[bootp + 0]		! print BOOTP address
	call	prnwrd
	mov	si,#crlf
	call	prnstr
#endif

! Check that we are running on the correct processor.

	mov	cl,#33
	mov	ax,#$FFFF		! the 8088 and 8086 will actually
	shl	ax,cl			! shift over by 33 bits
	jz	invcpu

	push	sp
	pop	ax
	cmp	ax,sp			! 286 or better?
	jne	invcpu

	pushf				! save original flags
	pushf
	pop	ax
	or	ax,#$7000		! set IOPL field in status word
	push	ax
	popf
	pushf
	pop	ax
	popf				! restore original flags
	test	ax,#$7000		! a 386+ lets us set the IOPL field
	jnz	cpuok

invcpu:	mov	si,#cpuerr		! print error message
	call	prnstr
	jmp	doerr1

! Check if the boot image header is correct.

cpuok:	les	bx,[header]
	mov	si,#bmerr		! prepare for correct error message
	seg	es
	mov	eax,BOOT_HD_MAGIC[bx]
	cmp	eax,#BOOT_MAGIC		! compare boot rom magic numbers
	jne	doerr1

	mov	si,#vmerr		! prepare for correct error message
	seg	es
	mov	al,BOOT_HD_LENGTH[bx]
	shr	al,#4
	and	al,#$0F
	cmp	al,#VENDOR_SIZE		! check vendor ID size
	jne	doerr1
	xor	di,di
dovmag:	mov	al,vmagic[di]		! check vendor ID
	or	al,al
	jz	chkmem			! vendor ID ok, continue
	seg	es
	cmp	al,BOOT_HD_VENDOR[bx+di]
	jne	doerr1
	inc	di
	jmp	dovmag
doerr1:	jmp	near doerr

! Get the size of the base memory from the BIOS. We need at least
! 616 kB.

chkmem:	int	$12			! get memory size in kB
	cmp	ax,#MIN_MEM
	mov	si,#memerr		! check if enough
	jb	doerr1
	mov	[avlmem],ax		! save it for later

! Get the size of extended memory from the BIOS. This is needed for computing
! the address of a loaded ramdisk image.

	mov	ah,#$88
	int	$15
	mov	[extmem],ax

! Looks good. Now find some memory where to put dynamic data which hasnt
! to be saved after this loader terminated. We use the memory space between
! the image end of this loader and the memory end of this loader.

	mov	si,#recerr
	mov	al,#VENDOR_BOOTL
	call	fndldr			! find load record for our own
	jc	doerr1			! loader module
	seg	es
	mov	edx,BOOT_LD_ADDR[di]
	mov	ebx,edx
	seg	es			! compute end of loader image
	add	edx,BOOT_LD_ILENGTH[di]
	seg	es			! compute end of loader memory
	add	ebx,BOOT_LD_MLENGTH[di]
	sub	ebx,edx			! compute length of free memory
	mov	si,#memerr
	cmp	ebx,#MAX_DYNAMIC_LEN
	jb	doerr1
	shr	ebx,#4			! convert size into paragraphs
	cmp	ebx,#$00001000
	jb	dynm1			! larger than 64kB
	mov	bx,#$1000
dynm1:	mov	[dynsiz],bx		! save size of dynamic memory area

	add	edx,#$0000000F		! round base of dynamic memory to
	shr	edx,#4			! nearest paragraph
#if defined(ASM_DEBUG) || defined(ASM_DEBUG_VERBOSE)
	mov	si,#dynmsg
	call	prnstr
	mov	ax,dx
	call	prnwrd			! lets see what we got here
	mov	si,#crlf
	call	prnstr
#endif

! We can now setup the pointers to all dynamic data structures. Note
! that the bootp memory block comes last so that we can use larger
! bootp records if required (and supported by the server), up until
! the end of the dynamic memory area.

	mov	word ptr [adrptr + 2],dx
	mov	word ptr [pthptr + 2],dx
	mov	word ptr [botptr + 2],dx
	mov	word ptr [adrptr + 0],#0
	mov	word ptr [pthptr + 0],#0 + MAX_ADDRS_LEN
	mov	word ptr [botptr + 0],#0 + MAX_ADDRS_LEN + MAX_PATH_LEN

! Determine the kernel setup entry address. This is used to actually call
! the kernel lateron.

	mov	si,#recerr
	mov	al,#VENDOR_SETUP
	call	fndldr			! find load record for kernel
	jc	doerr4			! setup module
	seg	es
	mov	al,BOOT_LD_FLAGS[di]		! get load record flags
	test	al,#BOOT_FLAG_B0 + BOOT_FLAG_B1	! we are presently unable
	jnz	doerr4				! to handle relative addrs
	seg	es
	mov	eax,BOOT_LD_ADDR[di]	! get base address of setup module
	test	eax,#$FFF0000F		! address must be reachable and on
	jnz	doerr4			! paragraph boundary
	shr	eax,#4			! determine segment of setup code
	mov	word ptr [kernel + 0],#0
	mov	word ptr [kernel + 2],ax
#if defined(ASM_DEBUG) || defined(ASM_DEBUG_VERBOSE)
	push	ax
	mov	si,#kermsg
	call	prnstr
	pop	ax
	call	prnwrd			! let the user know
	mov	al,#$3A
	call	prnchr
	xor	ax,ax
	call	prnwrd
	mov	si,#crlf
	call	prnstr
#endif
	jmp	chkkrn
doerr4:	jmp	near doerr

! Now check that the kernel has been loaded at the correct place and that
! the setup code has a correct header. Also take care of preloaded ramdisk
! images. All these features require Linux version 1.3.73 or above.

chkkrn:	mov	si,#recerr
	mov	al,#VENDOR_KERNEL
	call	fndldr			! find load record for kernel image
	jc	doerr4
	seg	es
	mov	al,BOOT_LD_FLAGS[di]		! get load record flags
	test	al,#BOOT_FLAG_B0 + BOOT_FLAG_B1	! we are presently unable
	jnz	doerr4				! to handle relative addrs
	mov	si,#exterr
	seg	es
	mov	eax,BOOT_LD_ADDR[di]	! get base address of setup module
	cmp	eax,#KERNEL_LDADDR	! kernel has to be loaded into high
	jne	doerr4			! memory

	mov	si,#seterr
	les	di,[kernel]		! get address of kernel setup header
	seg	es			! check magic number
	cmp	dword ptr SU_MAGIC_ADDR[di],#SU_MAGIC
	jne	doerr4
	mov	si,#oldkrn
	seg	es			! check version number
	cmp	word ptr SU_VERSION_ADDR[di],#SU_VERSION
	jb	doerr4
	call	fndrd			! Check for pre-loaded ramdisk image
	or	si,si
	jnz	doerr4

! Before we can use the memory area of the boot rom we have to secure
! all interesting data, e.g. the BOOTP record. First check to see if
! we got a valid BOOTP reply here, and copy it to the dynamic memory area.

	cld
	les	di,[bootp]
	seg	es
	mov	al,BOOTP_OP[di]		! op code must indicate reply
	cmp	al,#BOOTP_REPLY
	jne	dobot2			! it isnt
	seg	es
	mov	eax,BOOTP_VEND[di]
	cmp	eax,#BOOTP_MAGIC_RFC
	je	dobot3
dobot2:	mov	si,#noberr		! we have to have a valid BOOTP
	jmp	doerr7			! record

dobot3:	push	ds
	les	di,[botptr]
	lds	si,[bootp]
	mov	bx,si
	xor	cx,cx
	add	si,#BOOTP_VEND + BOOTP_MAGIC_LEN
dobot4:	lodsb
	cmp	al,#BOOTP_RFC_NOP	! handle NOP tag
	jne	dobot5
	inc	cx
	cmp	cx,#16			! more than 16 NOP tags is VERY unusual
	jb	dobot4			! so the bootp record maybe broken
doerr6:	pop	ds
	mov	si,#btperr
doerr7:	jmp	near doerr

dobot5:	cmp	al,#BOOTP_RFC_END	! handle END tag
	jne	dobot6
	mov	dx,si
	sub	dx,bx			! compute length of bootp record
	cmp	dx,#BOOTP_SIZE
	jae	dobot7
	mov	dx,#BOOTP_SIZE		! use minimum size
	jmp	dobot7

dobot6:	lodsb				! handle all other tags
	movzx	cx,al
	add	si,cx			! jump to next tag
	xor	cx,cx			! reset NOP counter
	jmp	dobot4			! proceed with next tag

dobot7:	xor	ecx,ecx
	mov	cx,es			! determine end of dynamic memory
	seg	cs
	movzx	eax,word ptr [dynsiz]
	add	eax,ecx
	shl	eax,#4
	shl	ecx,#4
	movzx	edi,di			! determine start of BOOTP area
	add	ecx,edi
	sub	eax,ecx			! compute number of bytes available
	cmp	eax,#$00010000		! check if we have more than 64kB
	jae	dobot8
	cmp	dx,ax			! check if BOOTP record fits into
	jae	doerr6			! dynamic memory area

dobot8:	mov	cx,dx
	mov	si,bx			! restore source pointer
	rep
	movsb				! save the bootp record
	pop	ds
	mov	[botlen],dx		! set length of bootp record

! At this point all dynamic data, which is only necessary for this boot
! loader, has been saved. Now we can go ahead and acquire memory for every-
! thing which must remain intact when we call the kernel. This is basically
! the stack and the command line. We set this static data area right at
! the end of all load records.

setmem:	mov	bx,#BOOT_LD_ILENGTH	! find the highest reserved memory
	call	fndmem			! block
#if defined(ASM_DEBUG) || defined(ASM_DEBUG_VERBOSE)
	mov	si,#stamsg
	call	prnstr
	mov	ax,dx
	call	prnwrd			! lets see what we got here
	mov	si,#crlf
	call	prnstr
#endif
	mov	si,#recerr
	mov	bx,dx			! BX - base of free memory area
	mov	al,#VENDOR_CMDL
	call	fndldr			! find load record for cmd line
	jc	doerr			! unable to find record
	seg	es
	mov	eax,BOOT_LD_MLENGTH[di]	! get memory length of cmd line
	cmp	eax,#CMDL_SIZE		! command line is not allowed to be
	ja	doerr			! longer than 2 kB
	mov	[cmdmax],ax		! save maximum size of command line
	add	ax,#STACK_SIZE		! compute size of static memory
	jc	doerr

	mov	si,#memerr
	add	ax,#$000F		! round up to nearest segment
	rcr	ax,#1
	shr	ax,#3			! convert it into segment number
	mov	dx,ax			! save size for later use
	add	ax,bx			! add it to base segment
	add	ax,#$003F		! round to nearest kB
	rcr	ax,#1
	shr	ax,#5			! convert it into kB
	cmp	ax,[avlmem]
	jb	setcmd			! we have enough memory

doerr:	call	prnstr			! in case of error return to the
	mov	si,oldSI		! boot rom with registers set
	mov	di,oldDI		! correctly
	mov	bp,oldBP
	mov	es,oldES
	mov	ds,oldDS
	retf

! We can now set the command line pointer. This requires some conversion
! because the kernel expects the command line to be relative to the floppy
! boot loader segment.
! BX contains the first segment usable for static memory, and DX contains
! the required size of static memory in paragraphs.

setcmd:	mov	si,#recerr
	mov	al,#VENDOR_INIT
	call	fndldr			! find load record for the INIT segment
	jc	doerr			! isnt there?
	seg	es
	mov	al,BOOT_LD_FLAGS[di]		! get load record flags
	test	al,#BOOT_FLAG_B0 + BOOT_FLAG_B1	! we are presently unable
	jnz	doerr				! to handle relative addrs
	seg	es
	mov	eax,BOOT_LD_ADDR[di]	! get base address of INIT segment
	test	eax,#$FFF0000F		! must be on a segment boundary in lower
	jnz	doerr			! memory
	shr	eax,#4			! compute start segment of INIT segment
	mov	word ptr [flopptr + 0],#0
	mov	word ptr [flopptr + 2],ax

	mov	si,#memerr
	sub	bx,ax			! compute offset to static area in
	jc	doerr			! paragraphs
	add	dx,bx			! the complete segment should not exceed
	jc	doerr			! 64 kB
	test	dx,#$F000
	jnz	doerr
	mov	[stasiz],dx
	mov	dx,ax
	shl	bx,#4			! convert segment offset into byte offset
	add	bx,#STACK_SIZE		! save the command line pointer
	mov	word ptr [cmdptr + 0],bx
	mov	word ptr [cmdptr + 2],dx

#if defined(ASM_DEBUG) || defined(ASM_DEBUG_VERBOSE)
	mov	si,#cmdmsg
	call	prnstr
	mov	ax,dx
	call	prnwrd
	mov	al,#$3A
	call	prnchr			! show the new command line pointer
	mov	ax,bx			! on the console
	call	prnwrd
	mov	si,#crlf
	call	prnstr
#endif

! Now set the stack. From this point on there is no return to the boot rom.

	cli				! disable interrupts during stack setup
	sub	bx,#2
	mov	ss,dx
	mov	sp,bx			! set the stack
	sti				! it is safe again

! At this point the memory layout should look like this, assuming default
! values for all load records set by mknbi:
!
! 0x020000-0x0201FF    0.5 kB	boot image header
! 0x020200-0x0209FF    2.0 kB	default command line
! 0x021000-0x02xxxx    x.x kB	primary boot loader - this program
! 0x02xxxx-0x02xxxx    0.5 kB	dynamic memory - IP address string
! 0x02xxxx-0x02xxxx    1.0 kB	dynamic memory - path name
! 0x02xxxx-0x024FFF    x.x kB   dynamic memory - BOOTP record
! 0x030000-0x0301FF    0.5 kB	linux floppy boot sector
! 0x030200-0x03xxxx    x.x kB	kernel setup
! 0x03xxxx-0x03xxxx    4.0 kB	static memory - stack
! 0x03xxxx-0x03xxxx    2.0 kB   static memory - kernel command line
! 0x100000-0x1FFFFF 1024.0 kB	kernel
!
! The remaining part of this program just deals with setting up the command
! line in the static memory area. First copy it to its rightful place and
! setup the pointers so that the kernel will find the new command line.

	push	ds
	xor	dx,dx			! DX contains length of command line
	mov	bp,[cmdmax]		! BP contains maximum length of cmd line
	sub	bp,#2			! keep some space for trailing 0
	mov	al,#VENDOR_CMDL
	call	fndldr			! find load record for cmd line
	mov	ax,es
	mov	si,di
	les	di,[cmdptr]		! load destination pointer
	mov	ds,ax
	jc	nocmd1			! unable to find command line
	mov	al,BOOT_LD_FLAGS[si]		! get load record flags
	test	al,#BOOT_FLAG_B0 + BOOT_FLAG_B1	! we are presently unable
	jnz	nocmd1				! to handle relative addrs

	mov	eax,BOOT_LD_ADDR[si]
	mov	si,ax
	shr	eax,#4			! compute segment
	test	eax,#$FFFF0000
	jnz	nocmd1			! segment is unreachable
	and	si,#$000F
	mov	ds,ax			! DS:SI points to old command line

	call	strlen			! get length of string
	jcxz	nocmd1
	cmp	cx,bp			! check length of command line
	jbe	stcmd1
	mov	cx,bp
stcmd1:	mov	dx,cx			! save length for later
	cld
	rep
	movsb				! move it
nocmd1:	xor	al,al
	stosb				! terminate the string with 0
	pop	ds
	mov	[cmdlen],dx		! save command line length

	les	di,[flopptr]
	mov	ax,word ptr [cmdptr + 0]
	seg	es			! setup command line descriptor
	mov	word ptr CL_MAGIC_ADDR[di],#CL_MAGIC
	seg	es
	mov	word ptr CL_OFFSET[di],ax

#ifdef ASM_DEBUG_VERBOSE
	mov	ax,dx
	call	prnwrd
	mov	al,#$20
	call	prnchr
	push	ds
	lds	si,[cmdptr]		! print the command line so far
	call	prnstr
	pop	ds
	mov	si,#crlf
	call	prnstr
#endif

! Now determine the name of the root directory on the server by looking
! into the BOOTP record. If a boot image file name was found, only leave
! the directory name by removing everything after the last slash. There
! is also the possibility to define a root directory name in the vendor
! area of the BOOTP record, so check for that also.

	push	ds
	mov	al,#BOOTP_RFC_FILE
	call	gettag			! check for boot image file name
	jcxz	bootp3
	mov	ax,es
	mov	si,di			! get boot image file name from BOOTP
	les	di,[pthptr]		! record
	mov	ds,ax
	jmp	bootp6

bootp3:	xor	bx,bx
	les	di,[pthptr]
	lds	si,[botptr]
	add	si,#BOOTP_FILE
	call	strlen			! determine length of file name
	jcxz	bootp2			! no file name, leave path untouched
bootp6:	mov	dx,#MAX_PATH_LEN - 2	! keep space in path name for trailing 0
	cmp	cx,dx
	jbe	bootp1
	mov	cx,dx			! get length of path name
bootp1:	mov	bx,cx			! save length for later
	cld
	rep
	movsb				! copy path name to its new location
	xor	al,al
	stosb				! save trailling 0
bootp2:	pop	ds

	xor	dx,dx
	or	bx,bx			! check if we have a file name
	jz	bootp4
	std
	les	di,[pthptr]
	mov	cx,bx
	add	di,bx			! point ES:DI to end of string
	mov	al,#$2F			! search backwards for last occurrence
	repne				! of a slash
	scasb
	jnz	bootp4			! not found -> no path
	seg	es
	mov	byte ptr [di + 1],#0	! replace the slash with 0
	mov	dx,cx
	inc	dx
bootp4:	mov	[pthlen],dx		! store length of directory name

	mov	al,#BOOTP_RFC_ROOT	! there is still another possibility:
	call	gettag			! get the directory name from vendor
	jcxz	bootp5			! field of bootp record
	mov	dx,cx
	push	ds
	cld
	mov	ax,es
	mov	si,di
	les	di,[pthptr]		! copy the directory name from the
	mov	ds,ax			! bootp record
	rep
	movsb
	xor	al,al
	stosb				! terminate string with 0
	pop	ds
	mov	[pthlen],dx

bootp5:
#ifdef ASM_DEBUG_VERBOSE
	mov	ax,[pthlen]
	call	prnwrd
	mov	al,#$20
	call	prnchr
	push	ds
	lds	si,[pthptr]		! print out the path
	call	prnstr
	pop	ds
	mov	si,#crlf
	call	prnstr
#endif

! Build the IP address string by looking into the bootp record.

	cld
	push	ds
	xor	dx,dx
	les	di,[adrptr]
	lds	si,[botptr]
	add	si,#BOOTP_YIADDR	! load my own IP address
	call	cvtadr			! put address into destination string
	mov	al,#$3A			! add ':'
	call	cvtchr
	add	si,#BOOTP_SIADDR - BOOTP_YIADDR
	call	cvtadr			! put server address into dest string
	mov	al,#$3A			! add ':'
	call	cvtchr
	pop	ds

	push	es
	push	di
	mov	al,#BOOTP_RFC_GWY	! we have to take the gateway address
	call	gettag			! from the vendor info area
	mov	ax,es
	mov	si,di
	pop	di
	pop	es
	jcxz	noadd1			! no netmask found
	push	ds
	mov	ds,ax
	call	cvtadr			! put gateway address into dest string
	pop	ds
noadd1:	mov	al,#$3A			! add ':'
	call	cvtchr

	push	es
	push	di
	mov	al,#BOOTP_RFC_MSK	! we have to take the netmask from the
	call	gettag			! vendor info area
	mov	ax,es
	mov	si,di
	pop	di
	pop	es
	jcxz	noadd2			! no netmask found
	push	ds
	mov	ds,ax
	call	cvtadr			! put netmask address into dest string
	pop	ds
noadd2:	mov	al,#$3A			! add ':'
	call	cvtchr

	push	es
	push	di
	mov	al,#BOOTP_RFC_HNAME	! get the host name from the vendor
	call	gettag			! info area
	mov	ax,es
	mov	si,di
	pop	di
	pop	es
	jcxz	noadd3			! no host name found
	push	ds
	mov	ds,ax
	add	dx,cx			! update string length
	rep
	movsb				! move host name into dest string
	pop	ds

noadd3:	push	es
	push	di
	mov	al,#BOOTP_RFC_RDEV	! get the name for the network device
	call	gettag			! by which to reach the root FS
	mov	ax,es
	mov	si,di
	pop	di
	pop	es
	jcxz	noaddr			! no device name found
	push	ds
	mov	ds,ax
	mov	al,#$3A			! only add ':' if there is a device name
	call	cvtchr
	add	dx,cx			! update string length
	rep
	movsb				! move device name into dest string
	pop	ds

noaddr:	xor	al,al			! terminate string with 0
	stosb
	mov	[adrlen],dx		! save string length

#ifdef ASM_DEBUG_VERBOSE
	mov	ax,dx
	call	prnwrd
	mov	al,#$20
	call	prnchr
	push	ds
	lds	si,[adrptr]		! let the user know
	call	prnstr
	pop	ds
	mov	si,#crlf
	call	prnstr
#endif

! Now we can create the final command line which can be passed to the
! kernel. This includes adding any additional arguments from the BOOTP
! record.

	mov	si,#rtstr
	mov	cx,#rtend - rtstr
	les	di,[pthptr]
	mov	ax,[pthlen]		! first handle root directory name
	or	ax,ax
	jnz	setup1
	mov	es,ax
	mov	di,ax
setup1:	call	procv			! setup command line

	mov	si,#adrstr
	mov	cx,#adrend - adrstr
	les	di,[adrptr]
	mov	ax,[adrlen]		! then handle the IP address string
	or	ax,ax
	jnz	setup2
	mov	es,ax
	mov	di,ax
setup2:	call	procv			! setup command line
	call	vparms			! process kernel parameters in vendor
					! extension tags
	mov	al,#BOOTP_RFC_CMDAD	! check for application vendor tag
	call	gettag
	jcxz	dovga			! tag not available
	mov	si,#nulstr
	call	addarg			! add string to end of command line
dovga:	call	vga			! process "vga=" argument

#ifdef ASM_DEBUG_VERBOSE
	mov	ax,[cmdlen]
	call	prnwrd
	mov	al,#$20
	call	prnchr
	push	ds
	lds	si,[cmdptr]		! Print the new command line for
	call	prnstr			! debugging
	pop	ds
#endif

! Setup the kernel header and finally call the kernel.

	les	di,[kernel]
	mov	ax,[stasiz]
	shl	ax,#4
	add	ax,#SECTSIZE - 1
	and	ax,#~(SECTSIZE - 1)
	seg	es			! set total size of setup code
	mov	SU_MOVE_SIZE[di],ax
	add	ax,#LINUX_HEAP
	seg	es			! set heap pointer to heap end
	mov	SU_HEAP[di],ax
	seg	es			! this is "mknbi"
	mov	byte ptr SU_TYPE_OF_LOADER[di],#SU_MY_TYPE_OF_LOADER
	seg	es			! can use heap, loaded high
	mov	byte ptr SU_LOAD_FLAGS[di],#SU_MY_LOAD_FLAGS
	seg	es			! kernel always loaded into high memory
	mov	dword ptr SU_32_START[di],#KERNEL_LDADDR
	mov	eax,[ramdisk]
	seg	es			! tell the kernel about the ramdisk
	mov	SU_RAMDISK_START[di],eax
	mov	eax,[rdsize]
	seg	es
	mov	SU_RAMDISK_SIZE[di],eax

#ifdef ASM_FREEZE_AFTER_INIT
lop1:	jmp	lop1			! This is a feature, not a bug...
#endif
#ifdef ASM_KEYPRESS_AFTER_INIT
	xor	ah,ah
	int	$16
#endif
	jmp	far ptr [kernel]


!
!====================================================================
!
! Convert an IP address pointed to by DS:SI into a string pointed
! to by ES:DI. The length of the string gets accumulated in DX.
! ES:DI points to first character after string.
!
! Changed registers: AX, DX, DI, ES
!
cvtadr:	push	bx
	mov	ax,word ptr [si + 0]
	or	ax,word ptr [si + 2]	! if address is 0.0.0.0, dont
	jz	cvtad9			! convert it

	xor	bx,bx
cvtad1:	mov	al,byte ptr [si + bx]
	call	cvtdec			! convert each byte in turn into a
	cmp	bx,#3			! decimal number
	jae	cvtad9			! abort if at last byte of address
	mov	al,#$2E			! add '.'
	call	cvtchr
	inc	bx			! proceed with next byte
	jmp	cvtad1

cvtad9:	pop	bx
	ret


!
!====================================================================
!
! Convert byte in AL into a decimal number string pointed to by
! ES:DI. The length of the string gets accumulated in DX. Return
! a pointer to first character after string in ES:DI.
!
! Changed registers: AX, DX, DI, ES
!
cvtdec:	push	cx
	xor	ah,ah
	xor	ch,ch
	mov	cl,#100
	div	cl			! get 100th
	or	ch,al			! dont print if zero
	jz	cvtd1
	call	cvtd7			! convert digit into ascii character
cvtd1:	mov	al,ah
	xor	ah,ah
	mov	cl,#10
	div	cl			! get 10th
	or	ch,al			! dont print if zero
	jz	cvtd2
	call	cvtd7			! convert digit into ascii character
cvtd2:	mov	al,ah			! get 1th
	pop	cx
cvtd7:	add	al,#$30			! convert digit into ascii character
cvtchr:	cmp	dx,#MAX_ADDRS_LEN
	jae	cvtch9
	stosb
	inc	dx
cvtch9:	ret


!
!====================================================================
!
! Scan all load records, and find the highest used memory address
! in base memory, round it up to the nearest segment boundary and
! return it in DX. The offset to the length value in the load records
! is given in BX. Ignore all memory segments that have been loaded
! into extended memory!
!
! Changed registers: AX, DX, DI, ES
!
fndmem:	les	di,[header]
	mov	dx,di
	add	dx,#$000F
	rcr	dx,#1			! round up to nearest segment
	shr	dx,#3
	mov	ax,es			! convert address of header into
	add	dx,ax			! segment representation, and add
	jc	fndm1			! the size of the header in order to
	add	dx,#BOOT_SIZE / 16	! get a starting value for the highest
	jnc	fndm2			! memory segment.
fndm1:	xor	dx,dx			! DX contains the highest block found

fndm2:	seg	es
	mov	al,BOOT_HD_LENGTH[di]	! get length of image header
fndm3:	call	getlen
	add	di,ax			! get the pointer to first load record
	seg	es
	mov	al,BOOT_LD_FLAGS[di]		! get load record flags
	test	al,#BOOT_FLAG_B0 + BOOT_FLAG_B1	! we are presently unable
	jnz	fndm4				! to handle relative addrs
	seg	es
	mov	eax,BOOT_LD_ADDR[di]	! get the load address
	seg	es
	add	eax,[di + bx]		! add the length
	add	eax,#$0000000F
	rcr	eax,#1			! round up and convert to nearest
	shr	eax,#3			! paragraph
	test	eax,#$FFFF0000		! only allow addresses in lower 1MB
	jnz	fndm4
	cmp	ax,dx			! is new address higher than old one?
	jbe	fndm4
	mov	dx,ax			! yes, save it

fndm4:	seg	es
	mov	al,BOOT_LD_FLAGS[di]	! ok, thats it, now check if thats the
	test	al,#BOOT_FLAG_EOF	! last record
	seg	es
	mov	al,BOOT_LD_LENGTH[di]	! get length of load record
	jz	fndm3
	ret


!
!====================================================================
!
! Find a load record in the boot header. The ID number of the load
! record is in AL, and ES:DI points to requested load record. It
! returns the carry flag set if load record not found.
!
! Changed registers: AX, DI, ES
!
fndldr:	push	cx
	mov	ch,al
	les	di,[header]		! examine boot image header
	seg	es
	mov	al,BOOT_HD_LENGTH[di]	! get length of image header
	call	getlen
	add	di,ax			! get the pointer to first load record
fndl1:	seg	es
	cmp	ch,BOOT_LD_TAG1[di]	! is it the desired one ?
	je	fndl3
	seg	es			! no, so check if its the last record
	test	byte ptr BOOT_LD_FLAGS[di],#BOOT_FLAG_EOF
	jnz	fndl2
	seg	es
	mov 	al,BOOT_LD_LENGTH[di]	! no, get the address of the next one
	call	getlen
	add	di,ax
	jmp	fndl1

fndl2:	stc				! couldnt find the desired record
fndl3:	pop	cx
	ret


!
!====================================================================
!
! Compute the length of a load record address from a length byte
! in AL. Return the offset in AX.
!
! Changed registers: AX
!
getlen:	mov 	ah,al
	shr	ah,#4
	and	ax,#$0f0f		! compute the total length in
	add	al,ah			! bytes from the length of the
	xor	ah,ah			! record and that of the vendor
	shl	ax,#2			! information.
	ret


!
!====================================================================
!
! Find a vendor tag in the BOOTP record. The tag number is in AL.
! It returns the pointer to the tag string in ES:DI and the length
! of the string in CX. CX is zero if tag not found.
!
! Changed registers: AX, CX, DI, ES
!
gettag:	cld
	push	si
	push	bx
	les	di,[botptr]
	mov	si,di
	add	si,[botlen]		! let ES:SI point to last bootp byte
	add	di,#BOOTP_VEND + BOOTP_MAGIC_LEN

	mov	ah,al			! save tag number
	xor	ch,ch
	xor	bx,bx
gett1:	seg	es
	mov	al,[di]			! load current vendor tag
	inc	di
	cmp	al,#BOOTP_RFC_NOP	! NOP tag --> continue
	je	gett1
	cmp	al,#BOOTP_RFC_END	! END tag --> abort
	je	gett3
	seg	es
	mov	cl,[di]			! get length of current tag
	inc	di
	cmp	al,ah			! we got our tag
	je	gett9
	or	bx,bx			! check for overload flag only once
	jnz	gett2
	cmp	al,#BOOTP_RFC_OVRLD	! check for option overload
	jne	gett2
	seg	es
	mov	bl,[di]			! get option overload flag
	inc	bh			! set to not accept another overload
gett2:	add	di,cx			! proceed with next tag
	cmp	di,si
	jb	gett1			! got at end of bootp record

gett3:	test	bl,#BOOTP_OVR_FILE
	jz	gett4
	les	di,[botptr]		! scan file section of BOOTP record
	add	di,#BOOTP_FILE
	mov	si,di
	add	si,#BOOTP_VEND - BOOTP_FILE
	and	bl,#~BOOTP_OVR_FILE
	jmp	gett1

gett4:	test	bl,#BOOTP_OVR_SNAME
	jz	gett8
	les	di,[botptr]		! scan server name section
	add	di,#BOOTP_SNAME
	mov	si,di
	add	si,#BOOTP_FILE - BOOTP_SNAME
	and	bl,#~BOOTP_OVR_SNAME
	jmp	gett1

gett8:	xor	cx,cx
gett9:	pop	bx
	pop	si
	ret


!
!====================================================================
!
! Process an argument on the command line. This routine handles
! basically three cases:
!   - the argument value is "rom", which means to use the value
!     delivered by the boot rom
!   - the argument value is "kernel", which means to just remove
!     the argument from the command line and then let the kernel
!     use its defaults
!   - the argument value contains neither "rom" nor "kernel", so
!     dont touch it.
!
! The name of the argument to look for in the command line is given
! in DS:SI with its length in CX. The boot rom value is pointed to
! by ES:DI. If ES:DI is the NULL pointer, no boot rom value is known.
!
! Changed registers: AX, BX, CX, DX, SI, DI, ES
!
procv:	push	bp
	push	es
	push	di
	les	di,[cmdptr]
	mov	dx,di			! load pointer to command line and
	call	strstr			! find option name
	jcxz	procv3			! option not found in cmd line
	mov	bx,di			!    ES:BX points to option name
	add	di,cx			!    ES:DI points to argument
	add	dx,[cmdlen]		!    DX    contains remaining cmd line
	sub	dx,di
	jle	procv2			! length of remaining cmd line is zero
	mov	cx,dx
	mov	dx,di
	mov	al,#$20			! find end of argument (terminated by
	repne				! blank) or end of string (e.g. CX=0)
	scasb
	jcxz	procv1			! argument goes to end of string
	dec	di
procv1:	sub	di,dx
	xchg	dx,di			! compute length of argument in DX
	jg	procv4			! argument length is greater than zero

procv2:	mov	dx,si
	mov	di,bx			! if we got here, we found the option
	call	rmarg			! name but no argument, so just remove
	mov	si,dx			! the option from the cmd line.
procv3:	xor	dx,dx
procv4:	pop	bp
	pop	cx			! get pointer to boot rom value

! At this point the registers are allocated as follows:
!      DS:SI  points to name of option
!      CX:BP  points to boot rom value
!      ES:BX  points to option name in command line
!      ES:DI  points to argument in command line
!      DX     contains length of argument; if 0, BX and DI are invalid
!
! After all that we have a couple of choices for constructing the kernel
! command line:
!
! argument not given in command line:
!   -  if there is no value from the boot rom, just do nothing and let
!      the kernel decide
!   -  if there is a value from the boot rom, add a new argument at the
!      end of the command line using the boot rom value
!
! argument given in command line:
!   -  if the value of the argument is not empty and not "rom" or "kernel",
!      just do nothing
!   -  if the value of the argument is "rom", then remove the argument
!      from the command line and add a new one with the boot rom value, but
!      only if a boot rom value is available
!   -  if the value of the argument is "kernel", then remove the argument
!      from the command line and do nothing, thus letting the kernel decide

	or	dx,dx			! check if argument found in command
	jnz	isarg			! line

	push	ds
	mov	ax,cx
	mov	ds,ax
	or	ax,bp
	jz	procv5
	seg	ds			! if no boot rom value is available,
	mov	al,[bp]			! e.g. the string pointer is NULL or
	or	al,al			! the string length is zero, then
	jz	procv5			! just return and do nothing.
	pop	ds
	jmp	procv6
procv5:	pop	ds
	jmp	procv9

procv6:	mov	es,cx
	mov	di,bp			! if not, add new option with boot rom
	call	addarg			! argument to end of command line
procv9:	pop	bp
	ret

! Original command line contains the option we are looking for. Check whether
! the argument is one of the strings "rom" or "kernel"

isarg:	cld
	push	cx
	push	si
	mov	ax,di
	mov	cx,dx
	cmp	cx,#romend - romstr
	jne	norom
	mov	si,#romstr
	repe
	cmpsb				! compare for "rom"
	mov	di,ax
	jz	dorom
norom:	mov	cx,dx
	cmp	cx,#kerend - kerstr
	jne	nokern
	mov	si,#kerstr
	repe
	cmpsb				! compare for "kernel"
	mov	di,ax
	jz	dokern
nokern:	pop	si
	pop	cx			! not found --> do nothing
	jmp	procv9

! the argument is "rom", so remove the existing argument from the
! command line, and add the boot rom path

dorom:	mov	di,bx
	call	rmarg			! remove old option containing
	pop	si			! "rom" argument
	pop	es
	mov	di,bp

	mov	ax,es
	or	ax,di
	jz	procv9
	seg	es			! if no boot rom value is available,
	mov	al,[di]			! e.g. the string pointer is NULL or
	or	al,al			! the string length is zero, then
	jz	procv9			! just return and do nothing.
	call	addarg			! add new option with boot rom value
	jmp	procv9			! as argument

! the argument is "kernel", so just remove the existing argument from
! the command line and let the kernel decide by itself.

dokern:	mov	di,bx
	call	rmarg			! remove old option containing
	pop	si			! "kernel" argument
	pop	cx
	jmp	procv9


!
!====================================================================
!
! Add a new option to the end of the command line. DS:SI points to
! name of option, and ES:DI points to argument value. Return ES:DI
! pointing to the trailing 0 of the command line, and the new length
! of the command line in AX.
!
! Changed registers: AX, SI, DI, ES
!
addarg:	push	cx
	push	dx
	call	strlen			! first compute the new length of
	mov	dx,cx			! the command line. if the new
	xchg	si,di			! argument will not fit into the
	push	ds			! memory provided, then just return
	mov	ax,es			! and leave the command line
	mov	ds,ax			! unchanged.
	call	strlen
	pop	ds
	xchg	si,di
	add	cx,dx
	pop	dx
	add	cx,[cmdlen]
	inc	cx
	cmp	cx,[cmdmax]
	pop	cx
	jae	addar9

	cld
	push	bx
	push	ds
	push	di
	push	es
	les	di,[cmdptr]
	mov	bx,di
	add	di,[cmdlen]
	mov	al,#$20			! add a blank
	stosb
addar1:	lodsb				! add the option name
	or	al,al
	jz	addar2
	stosb
	jmp	addar1
addar2:	pop	ds
	pop	si			! load the argument pointer
addar3:	lodsb				! add the argument
	or	al,al
	jz	addar4
	stosb
	jmp	addar3
addar4:	pop	ds			! restore DS and add trailing 0
	xor	al,al
	stosb
	dec	di
	mov	ax,di
	sub	ax,bx
	mov	[cmdlen],ax		! adjust length of command line
	pop	bx
addar9:	ret


!
!====================================================================
!
! Remove an argument pointed to by ES:DI from the command line.
! Return ES:DI pointing to the trailing 0 of the command line, and
! the new length of the command line in AX.
!
! Changed registers: AX, SI, DI, ES
!
rmarg:	cld
	push	bx
	push	cx
	push	dx
	push	ds
	lds	si,[cmdptr]
	mov	ax,ds			! get beginning of command line
	mov	es,ax
	mov	bx,si
	mov	si,di
	call	strlen			! get remaining length of cmd line
	mov	ax,#$20			! find end of argument in command line
	repne				! DS:SI points to first character after
	scasb				!       the argument string
	dec	di			! ES:DI points to start of argument
	xchg	si,di			! ES:BX points to start of cmd line
	or	cx,cx			! argument is not at the end of the
	jnz	rmarg2			! command line

rmarg1:	cmp	di,bx
	je	rmarg4			! empty command line
	dec	di			! remove leading blanks
	seg	es
	cmp	byte ptr [di],#$20
	je	rmarg1
	inc	di			! point to first blank after preceeding
	jmp	rmarg4			! argument and make it end of string

rmarg2:	lodsb
	cmp	al,#$20
	je	rmarg2			! remove trailing blanks
rmarg3:	or	al,al
	jz	rmarg1			! end of command line... remove all
	stosb				! trailing blanks. Otherwise copy
	lodsb				! the remainder of the command line
	jmp	rmarg3			! over the argument

rmarg4:	xor	al,al
	stosb				! add trailing 0
	pop	ds
	pop	dx
	pop	cx
	dec	di			! point to trailing 0
	mov	ax,di
	sub	ax,bx			! adjust length of command line
	mov	[cmdlen],ax
	pop	bx
	ret


!
!====================================================================
!
! Find a substring (DS:SI, length CX) in a string (ES:DI). Return
! the substring position in ES:DI. CX is unchanged if substring found,
! otherwise CX is 0.
!
! Changed registers: AX, CX
!
strstr:	push	bx
	push	dx
	cld				! DX contains length of substring
	mov	dx,cx			! ES:BX points to string
	mov	bx,di			! DS:SI points to substring

strst1:	seg	es
	mov	al,[bx]
	or	al,al			! if at end of string we could not find
	jz	strst2			! the substring
	mov	di,bx
	mov	cx,dx			! reset substring
	mov	ax,si
	repe				! compare substring at current string
	cmpsb				! position
	mov	si,ax
	jz	strst3			! found it!
	inc	bx			! compare cmd line at next char
	jmp	strst1

strst2:	xor	dx,dx			! could not find string
strst3:	mov	cx,dx			! set return values
	mov	di,bx
	pop	dx
	pop	bx
	ret


!
!====================================================================
!
! Determine length of string (DS:SI), and return it in CX
!
! Changed registers: AL, CX
!
strlen: push	si
	xor	cx,cx
	cld
strl1:	lodsb
	or	al,al
	jz	strl3
	cmp	al,#$20			! character should be printable
	jl	strl2
	cmp	al,#$7E
	ja	strl2
	inc	cx			! count character
	jmp	strl1
strl2:	xor	cx,cx			! return 0 length if string invalid
strl3:	pop	si
	ret


!
!====================================================================
!
! Print a string in DS:SI onto the console
!
! Changed registers: AL
!
prnstr:	push	si
	cld
prns1:	lodsb				! loop over all characters of
	or	al,al			! string
	jz	prns2
	call	prnchr			! print character
	jmp	prns1
prns2:	pop	si
	ret


!
!====================================================================
!
! Print hexadecimal values (in AX or AL) or characters onto the console
!
! Changed registers: AX
!
prndwd:	push	eax
	shr	eax,#16
	call	prnwrd			! print the upper word
	pop	eax
prnwrd:	push	ax
	mov	al,ah
	call	prnbyt			! print the upper byte
	pop	ax
prnbyt: push	ax
	shr	al,#4			! prepare upper nibble
	call	prnnib			! print it
	pop	ax
prnnib:	and	al,#$0F			! prepare lower nibble
	add	al,#$30
	cmp	al,#$39			! convert it into hex
	jle	prnchr
	add	al,#$07
prnchr:	push	bx
	mov	ah,#$0E			! print it
	mov	bx,#$0007
	int	$10
	pop	bx
	ret


!
!====================================================================
!
! Check whether a ramdisk image is attached to the current netboot image;
! if a ramdisk is found, determine load address and size. If an error
! occurred, SI contains a pointer to the error message.
!
! Changed registers: EAX, EBX, ECX, EDX, SI, ES, DI
!
fndrd:	xor	si,si
	mov	al,#VENDOR_RAMDISK
	call	fndldr			! find load record for ramdisk image
	jc	rdret2			! no ramdisk image found
	mov	si,#recerr
	seg	es
	mov	eax,BOOT_LD_ILENGTH[di]	! determine size of ramdisk image
	mov	[rdsize],eax

	seg	es
	mov	bl,BOOT_LD_FLAGS[di]	! get load record flags
	test	bl,#BOOT_FLAG_B0
	jnz	rdret2			! error: neither RD_FIX/AUTO nor RD_EOM
	movzx	eax,word ptr [extmem]	! determine end of physical memory
	test	bl,#BOOT_FLAG_B1
	jnz	rdeom			! ramdisk is at end of memory
	seg	es			! check if ramdisk image needs to be
	mov	bl,BOOT_LD_VENDOR[di]	! relocated to end of memory
	cmp	bl,#RD_AUTO
	jne	rdfix			! ramdisk is at fixed location
	cmp	ax,#$3C00		! there is more then 16MB
	jbe	rdeom
	mov	ax,#$3C00		! limit to 16MB for BIOS
rdeom:	add	eax,#$00000400		! add size of conventional memory
	shl	eax,#10			! multiply by 1024
	seg	es
	sub	eax,BOOT_LD_MLENGTH[di]	! compute start address of ramdisk
	seg	es
	mov	bl,BOOT_LD_VENDOR[di]	! check if ramdisk image needs to be
	cmp	bl,#RD_AUTO		! relocated to end of memory
	jne	rddone
	seg	es
	mov	ebx,BOOT_LD_ADDR[di]	! source address
	mov	ecx,[rdsize]
	call	moverd			! move ramdisk
	or	si,si
	jz	rddone			! check if error
rdret2:	jmp	rdret

rdfix:	seg	es
	mov	eax,BOOT_LD_ADDR[di]
rddone:	mov	[ramdisk],eax
#if defined(ASM_DEBUG) || defined(ASM_DEBUG_VERBOSE)
	push	eax
	mov	si,#rdmsg
	call	prnstr
	pop	eax
	call	prndwd			! let the user know
	mov	al,#$28
	call	prnchr
	mov	eax,[rdsize]
	call	prndwd
	mov	ax,#$29
	call	prnchr
	mov	si,#crlf
	call	prnstr
#endif
	xor	si,si
rdret:	ret


!
!====================================================================
!
! Move the ramdisk image from the location given in EAX to EBX with
! a size of ECX. If an error occurred, SI contains a pointer to the
! error message.
!
! Changed registers: ECX, EDX, SI
!
moverd:	push	eax			! save all registers that might
	push	ebx			! be mangled
	push	es
	push	di
	mov	word ptr [rd_dstb],bx	! enter destination address into gdt
	mov	word ptr [rd_srcb],ax	! enter source address into gdt
	inc	ecx			! int $15 moves words
	mov	edx,ecx
	shr	edx,#16
	shr	eax,#16
	shr	ebx,#16
	add	al,dl			! move from the top downwards
	add	bl,dl
	mov	byte ptr [rd_dstb + 2],bl
	mov	byte ptr [rd_dstb + 2],al
	shr	cx,#1
	jcxz	mvzero
	mov	ax,cs
	mov	es,ax
	mov	si,#rd_gdt
	mov	ax,#$8700		! have the BIOS move the memory
	int	$15
	mov	si,#moverr
	jc	mvret			! oops, something went wrong...
mvzero:	or	dl,dl
	jz	mvno64
mvloop:	dec	byte ptr [rd_srcb + 2]	! decrement source address
	dec	byte ptr [rd_dstb + 2]	! decrement dest address
	mov	cx,#$8000		! move in chunks of 64k/2 words
	mov	si,#rd_gdt
	mov	ax,#$8700
	int	$15
	mov	si,#moverr
	jc	mvret			! oops, something went wrong...
	dec	dl
	jnz	mvloop
mvno64:	xor	si,si
mvret:	pop	di			! restore all registers
	pop	es
	pop	ebx
	pop	eax
	ret


!
!====================================================================
!
! Handle any "vga=" argument that might be on the command line
!
! Changed registers: AX, BX, CX, DX, SI, DI, ES
!
vga:	mov	si,#vgastr
	mov	cx,#vgaend - vgastr
	les	di,[cmdptr]
	call	strstr			! ES:DI  = strstr(cmdptr,"vga=")
	jcxz	isret			! not found
	push	di
 	add	di,cx			! ES:DI += strlen("vga=")
	xor	bx,bx

	cld
	mov	ax,di
	mov	cx,#askend - askstr	! strncmp(ES:DI,"ASK",3)
	mov	si,#askstr
	repe
	cmpsb
	mov	di,ax
	jz	doask

	mov	cx,#extend - extstr	! strncmp(ES:DI,"EXTENDED",8)
	mov	si,#extstr
	repe
	cmpsb
	mov	di,ax
	jz	doext

	mov	cx,#nrmend - nrmstr	! strncmp(ES:DI,"NORMAL",6)
	mov	si,#nrmstr
	repe
	cmpsb
	mov	di,ax
	jz	donrm

	mov	cx,#hexend - hexstr	! strncmp(ES:DI,"$",2)
	mov	si,#hexstr
	repe
	cmpsb
	mov	di,ax
	jz	dohex
	jmp	dodez

isret:	ret

doask:	dec	bx			! ASK      == -3
doext:	dec	bx			! EXTENDED == -2
donrm:	dec	bx			! NORMAL   == -1
vgaset:	les	si,[flopptr]
	seg	es
	mov	word ptr VGA_OFFSET[si],bx

#if defined(ASM_DEBUG) || defined(ASM_DEBUG_VERBOSE)
	push	bx
	mov	si,#vgastr
	call	prnstr
	pop	ax
	call	prnwrd
	mov	si,#crlf
	call	prnstr
#endif

	les	di,[cmdptr]		! remove this argument from commandline
	pop	di
	call	rmarg
	jmp	vga			! handle next argument (if any)

dohex:	mov	cl,#4
	xor	dh,dh
	inc	di
hexloop:inc	di			! ES:DI++
	seg	es
	mov	dl,[di]
	sub	dl,#$30			! if (*ES:DI < '0') break
	jc	vgaset
	cmp	dl,#$09			! if (*ES:DI > '9') goto nodigit
	jbe	hexinc
	sub	dl,#$11			! if (*ES:DI < 'A') break
	jc	vgaset
	cmp	dl,#$05			! if (*ES:DI > 'F') goto nocapital
	jbe	hexxinc
	sub	dl,#$20			! if (*ES:DI < 'a') break
	jc	vgaset
	cmp	dl,#$05			! if (*ES:DI > 'f') break
	ja	vgaset
hexxinc:add	dl,#$0A
hexinc:	shl	bx,cl			! BX = 16*BX + digit
	add	bx,dx
	jmp	hexloop

vgasgn:	or	al,al			! if (sign) BX = -BX
	jz	vgaset
	neg	bx
	jmp	vgaset

dodez:	xor	al,al
	xor	dh,dh
	seg	es
	cmp	byte ptr [di],#$2D	! if (*ES:DI == '-') sign++
	jne	dezcnv
	inc	al			! AL = sign
dezlop:	inc	di			! ES:DI++
dezcnv:	seg	es
	mov	dl,[di]
	sub	dl,#$30			! if (*ES:DI < '0') break
	jc	vgasgn
	cmp	dl,#$09			! if (*ES:DI > '9') break
	ja	vgasgn
	add	bx,bx			! BX = 10*BX + *ES:DI - '0'
	mov	cx,bx
	shl	bx,#2
	add	bx,cx
	add	bx,dx
	jmp	dezlop


!
!====================================================================
!
! Process kernel parameters in vendor extension tags
!
! Changed registers: AX, BX, CX, DX, SI, DI, ES
!
vparms:	mov	al,#BOOTP_RFC_VID	! check for 'Eth' vendor ID
	call	gettag
	cmp	cx,#6			! there have to be at least 6 bytes
	jb	vperr1			! in this tag
	seg	es
	cmp	dword ptr [di],#BOOTP_VID_ID
	jne	vperr2

	mov	al,#BOOTP_RFC_VSEL	! get user selection
	call	gettag
	cmp	cx,#1
vperr1:	jb	vperr
	seg	es
	mov	al,byte ptr [di]	! get selected image descripton
	call	gettag
	jcxz	vperr

	cld
	seg	es
	mov	cl,byte ptr [di - 1]
	mov	bx,#6			! scan for parameter field
	mov	ax,#$3A			! count colons
vplop1:	repne
	scasb
	jcxz	vperr
	dec	bx
	jnz	vplop1

	mov	si,di
	repne				! scan for end of parameter field
	scasb
	seg	es
	cmp	byte ptr [di - 1],al
	jnz	vnocol
	dec	di
vnocol:	mov	cx,di			! compute length of parameter field
	sub	cx,si
	mov	bx,[cmdlen]
	add	bx,cx
	add	bx,#2
	cmp	bx,[cmdmax]
	jae	vperr
	dec	bx
	mov	ax,es
	les	di,[cmdptr]
	add	di,[cmdlen]
	push	ds
	mov	ds,ax
	mov	al,#$20			! add a blank
	stosb
vpcpy:	lodsb
	or	al,al
	jnz	vnonul
	sub	bx,cx
	jmp	vpeos

vnonul:	cmp	al,#$7E			! check for tilde character
	jnz	vnoesc
	dec	bx
	dec	cx
	jz	vpeos
	lodsb
	cmp	al,#$62			! '~b' -> '\\'
	jb	vnoesc
	jnz	vpnob
	mov	al,#$5C
	jmp	vnoesc

vperr2:	jmp	vperr

vpnob:	cmp	al,#$63			! '~c' -> ':'
	jnz	vnoesc
	mov	al,#$3A
vnoesc:	stosb
	loop	vpcpy			! loop over whole string

vpeos:	xor	al,al
	stosb				! add trailing 0
	pop	ds
	mov	[cmdlen],bx
vperr:	ret


!
!====================================================================
!
! String and constants definitions
!
! Startup signature
!
sigmsg:	.byte	$0D, $0A
	.ascii	"Linux Net Boot Image Loader "
	.ascii	VERSION
	.byte	$0D, $0A
	.ascii	"Copyright (C) 1996-2001 Gero Kuhlmann and M. Gutschke"
	.byte	$0D, $0A
	.ascii	COPYRIGHT
	.byte	$0D, $0A
crlf:	.byte	$0D, $0A
	.byte	0


! Boot image header vendor magic ID

vmagic:	.ascii	VENDOR_MAGIC
	.byte	0


! Possible option names for command line

rtstr:	.ascii	"nfsroot="
rtend:	.byte	0

adrstr:	.ascii	"nfsaddrs="
adrend:	.byte	0

vgastr:	.ascii	"vga="
vgaend:	.byte	0

askstr:	.ascii	"ASK"
askend:	.byte	0

extstr:	.ascii	"EXTENDED"
extend:	.byte	0

nrmstr:	.ascii	"NORMAL"
nrmend:	.byte	0

hexstr:	.ascii	"$"
hexend:	.byte	0

nulstr:	.byte	0


! Possible arguments for command line options

romstr:	.ascii	"rom"
romend:	.byte	0

kerstr:	.ascii	"kernel"
kerend:	.byte	0


! Error messages

cpuerr:	.ascii	"Not running on a 386+ processor"
	.byte	$0D, $0A
	.byte	0

memerr:	.ascii	"Not enough memory"
	.byte	$0D, $0A
	.byte	0

recerr:	.ascii	"Error in load record data"
	.byte	$0D, $0A
	.byte	0

noberr:	.ascii	"BOOTP/DHCP record missing"
	.byte	$0D, $0A, 0

btperr:	.ascii	"BOOTP/DHCP record too large"
	.byte	$0D, $0A
	.byte	0

seterr:	.ascii	"Error in kernel setup header"
	.byte	$0D, $0A
	.byte	0

exterr:	.ascii	"Kernel not loaded into high memory"
	.byte	$0D, $0A
	.byte	0

bmerr:	.ascii	"Invalid boot header magic number"
	.byte	$0D, $0A
	.byte	0

vmerr:	.ascii	"Invalid vendor magic ID"
	.byte	$0D, $0A
	.byte	0

oldkrn:	.ascii "Linux kernel too old"
	.byte	$0D, $0A
	.byte	0

moverr:	.ascii	"Failed to move ramdisk image to end of memory"
	.byte	$0D, $0A
	.byte	0


#if defined(ASM_DEBUG) || defined(ASM_DEBUG_VERBOSE)
! Some debug messages

ad1msg:	.ascii	"Header addr = "
	.byte	0

ad2msg:	.ascii	"BOOTP addr = "
	.byte	0

kermsg:	.ascii	"Kernel start: "
	.byte	0

rdmsg:	.ascii	"Ramdisk start: "
	.byte	0

cd32msg:.ascii	"Kernel image: "
	.byte	0

dynmsg:	.ascii	"Dynamic memory start: "
	.byte	0

stamsg:	.ascii	"Static memory start: "
	.byte	0

cmdmsg:	.ascii	"Address of command line: "
	.byte	0
#endif


!
!====================================================================
!
! Variable definitions
!
header:	.long	0			! pointer to boot header from boot rom
bootp:	.long	0			! pointer to bootp block from boot rom
kernel:	.long	0			! pointer to kernel entry

oldDS:	.word	0			! old DS from boot rom
oldES:	.word	0			! old ES from boot rom
oldBP:	.word	0			! old BP from boot rom
oldSI:	.word	0			! old SI from boot rom
oldDI:	.word	0			! old DI from boot rom

avlmem:	.word	0			! available amount of RAM in kB
extmem:	.word	0			! extended memory in kB
dynsiz:	.word	0			! size of dynamic memory in paragraphs
stasiz:	.word	0			! size of static memory in paragraphs

ramdisk:.long	0			! pointer to start of ramdisk (32bit)
rdsize:	.long	0			! size of ramdisk

cmdptr:	.long	0			! pointer to start of command line
cmdlen:	.word	0			! length of command line
cmdmax:	.word	0			! maximum length of command line

botptr:	.long	0			! pointer to bootp record
botlen:	.word	0			! length of bootp record

pthptr:	.long	0			! pointer to path string
pthlen:	.word	0			! length of path string

adrptr:	.long	0			! pointer to IP address string
adrlen:	.word	0			! length of IP address string

flopptr:.long	0			! pointer to floppy boot sector

rd_gdt:	.word	0,0,0,0
	.word	0,0,0,0
rd_src:	.word	$FFFF			! length
rd_srcb:.byte	0,0,0			! base
	.byte	$93			! typebyte
	.word	0			! limit16,base24 =0
rd_dst:	.word	$FFFF			! length
rd_dstb:.byte	0,0,0			! base
	.byte	$93			! typebyte
	.word	0			! limit16,base24 =0
	.word	0,0,0,0			! BIOS CS
	.word	0,0,0,0			! BIOS DS


!
!====================================================================
!
	end

