/*
 * memory.c  -  Memory allocation and private string handling
 *
 * Copyright (C) 1998-2001 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: memory.c,v 1.2 2001/08/27 08:40:16 gkminix Exp $
 */

#include <common.h>
#include <nblib.h>
#include "privlib.h"



/*
 * Allocate memory - with error handling
 */
voidstar nbmalloc(amount)
size_t amount;
{
  voidstar ptr;

  assert(amount > 0);
  if ((ptr = malloc(amount)) == NULL) {
	if (!quiet) perror(progname);
	exit(EXIT_MEMORY);
  }
  memset(ptr, 0, amount);
  return(ptr);
}



/*
 * Copy a string into a newly allocated memory block
 */
void copystr(dest, src)
char **dest;
char  *src;
{
  if (*dest != NULL)
	free(*dest);

  *dest = (char *)nbmalloc(strlen(src) + 1);
  strcpy(*dest, src);
}



/*
 * Compare a string in a binary buffer with a string. We can't use
 * strcmp() here, because characters might be represented with a different
 * size on the host system than on the x86 target system. It returns
 * TRUE, if the byte buffer contents and the string are identical.
 */
int bytecmp(str, buf, len)
char *str;
__u8 *buf;
size_t len;
{
  while (*str && len > 0) {
	if ((toascii(*str) & 0xff) != *buf)
		break;
	str++;
	buf++;
	len--;
  }
  return(len == 0);
}



/*
 * Copy a string into a byte array. We can't use strcpy() here because
 * characters might be represented differently on the system than on the
 * target x86 system. This does NOT copy the zero at the end of the
 * source string.
 */
void bytecpy(str, buf, len)
char *str;
__u8 *buf;
size_t len;
{
  while (*str && len > 0) {
	*buf = toascii(*str) & 0xff;
	buf++;
	str++;
	len--;
  }
}

